import os
import uuid

import boto3
from botocore.client import ClientError

from . import S3
from .s3_bucket_error import NumberRandomCharsException


class S3Object(S3):

    @staticmethod
    def create_name(object_suffix, num_random_chars=6):
        if num_random_chars < NumberRandomCharsException.min_num_chars or \
                num_random_chars > NumberRandomCharsException.max_num_chars:
            raise NumberRandomCharsException

        return ''.join([str(uuid.uuid4()).replace('-', '')[:num_random_chars], '-', object_suffix])

    @staticmethod
    def get_object_key(file_name, with_random_prefix=True, num_random_chars=6):
        if with_random_prefix:
            return S3Object.create_name(file_name, num_random_chars)
        else:
            return file_name

    @classmethod
    def upload(cls, bucket_name, file_dict, with_random_prefix=True, num_random_chars=6):
        if 'file_path' in file_dict:
            _, file_name = os.path.split(file_dict['file_path'])
        else:
            file_name = file_dict['file_name']

        object_key = cls.get_object_key(file_name, with_random_prefix, num_random_chars)

        if 'file_path' in file_dict:
            cls.resource.Object(bucket_name, object_key).upload_file(file_dict['file_path'])

        elif 'file_bytes' in file_dict:
            cls.resource.meta.client.put_object(Bucket=bucket_name, Key=object_key, Body=file_dict['file_bytes'])

        elif 'file_obj' in file_dict:
            cls.resource.meta.client.upload_fileobj(file_dict['file_obj'], bucket_name, object_key)

        else:
            raise NotImplementedError('The "file_dict" parameters should include file_obj or file_bytes or file_path')

        return object_key

    @classmethod
    def download(cls, bucket_name, object_key, file_dict=dict()):
        if 'file_path' in file_dict:
            return cls.resource.Object(bucket_name, object_key).download_file(file_dict['file_path'])

        elif 'file_obj' in file_dict:
            return cls.resource.meta.client.download_fileobj(bucket_name, object_key, file_dict['file_obj'])

        elif len(file_dict) == 0:
            # Get the file bytes
            return cls.resource.meta.client.get_object(Bucket=bucket_name, Key=object_key)['Body'].read()

        else:
            raise NotImplementedError(
                'The "file_dict" parameters should include file_obj, file_path, or empty dict to return the file bytes')

    @classmethod
    def delete(cls, bucket_name, object_key):
        return cls.resource.meta.client.delete_object(Bucket=bucket_name, Key=object_key)

    @classmethod
    def list_all(cls, bucket_name, as_generator):
        bucket = cls.resource.Bucket(bucket_name)
        objects = bucket.objects.all()

        if as_generator:
            return objects

        return list(objects)

    @classmethod
    def list_all_keys(cls, bucket_name, as_generator):
        objects = cls.list_all(bucket_name, as_generator=True)
        objects_keys = (obj.key for obj in objects)

        if as_generator:
            return objects_keys

        return list(objects_keys)

    @classmethod
    def is_available(cls, bucket_name, object_key):
        try:
            cls.resource.Object(bucket_name, object_key).load()

        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False

        else:
            return True

    @classmethod
    def num_objects(cls, bucket_name):
        return len(cls.list_all(bucket_name, False))

    @classmethod
    def generate_presigned_url(cls, bucket_name, object_name, expiration):
        """Generate a presigned URL to share an S3 object    :param bucket_name: string
        :param object_name: string
        :param expiration: Time in seconds for the presigned URL to remain valid
        :return: Presigned URL as string. If error, returns None.
        """

        response = cls.resource.meta.client.generate_presigned_url('get_object',
                                                                   Params={'Bucket': bucket_name,
                                                                           'Key': object_name},
                                                                   ExpiresIn=expiration)

        return response



class NumberRandomCharsException(Exception):
    min_num_chars = 0
    max_num_chars = 32

    def __init__(self):
        super().__init__(
            'The random characters are generated as a uuid4. Therefore, the selected number "num_random_chars" should be between {} to {} '.format(
                NumberRandomCharsException.min_num_chars, NumberRandomCharsException.max_num_chars))


class NumberCharsBucketNameException(Exception):
    min_num_chars = 3
    max_num_chars = 63

    def __init__(self, bucket_name, num_chars):
        super().__init__(
            'The number of characters in a bucket name should be between {} and {}. Your bucket name {} has {} chars only'.format(
                NumberCharsBucketNameException.min_num_chars, NumberCharsBucketNameException.max_num_chars,
                bucket_name, num_chars))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="aws-s3-resource",
    version="0.0.10",
    author="Quaking Aspen",
    author_email="info@quakingaspen.net",
    license='MIT',
    description="This library is to ease the operations done on AWS S3 buckets and objects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Quakingaspen-codehub/aws_s3_resource",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    platform=['Any'],
    python_requires='>=3.6',
    install_requires=['boto3', 'botocore']
)



from . import S3
from .s3_object import S3Object
from botocore.client import ClientError
import uuid
import os
import boto3

class S3Bucket(S3):
    
    @staticmethod
    def create_name(bucket_prefix, num_random_chars=6):
        
        if num_random_chars < NumberRandomCharsException.min_num_chars or \
           num_random_chars > NumberRandomCharsException.max_num_chars:
            
            raise NumberRandomCharsException
        
        bucket_name = ''.join([bucket_prefix, '-', str(uuid.uuid4()).replace('-', '')[:num_random_chars]])
        
        if len(bucket_name) < NumberCharsBucketNameException.min_num_chars or \
           len(bucket_name) > NumberCharsBucketNameException.max_num_chars:
            
            raise NumberCharsBucketNameException(bucket_name, len(bucket_name))
        
        return bucket_name
    
    @classmethod
    def create(cls, bucket_name):
        
        # Take the region from the config file
        session = boto3.session.Session()
        region = session.region_name

        # Create the bucket
        cls.resource.create_bucket(Bucket=bucket_name,
                          CreateBucketConfiguration={'LocationConstraint': region})
    
    @classmethod
    def create_with_random_name_suffix(cls, bucket_name, num_random_chars=6):
        while True:
            try:
                response = cls.create(bucket_name)
            except ClientError as e:
                if e.response['Error']['Code'] == 'BucketAlreadyExists':
                    bucket_name = cls.create_name(bucket_name)
                else:
                    raise e
            else:
                return bucket_name
    
    @classmethod
    def delete(cls, bucket_name):
        return cls.resource.Bucket(bucket_name).delete()
    
    @classmethod
    def list_all(cls, as_generator):
        buckets = cls.resource.buckets.all()
        
        if as_generator:
            return buckets
        
        return list(buckets)
    
    @classmethod
    def list_all_names(cls, as_generator):
        buckets = cls.list_all(as_generator=True)
        buckets_names = (bucket.name for bucket in buckets)
        
        if as_generator:
            return buckets_names 
        
        return list(buckets_names)
    
    @classmethod
    def is_available(cls, bucket_name):
        bucket = cls.resource.Bucket(bucket_name)
        return bucket in cls.resource.buckets.all()
    
    @classmethod
    def num_buckets(cls):
        return cls.list_all(False)

    @classmethod
    def download(cls, bucket_name, folder_name):
        # Create the root folder
        os.makedirs(folder_name)
        bucket = cls.resource.Bucket(bucket_name)

        # A set contains all created folders
        created_folders = set()

        for my_bucket_object in bucket.objects.all():

            # Folder or file path
            path = os.path.join(folder_name, my_bucket_object.key)

            # If the object name end with special character "/", it will definitely be a folder 
            if my_bucket_object.key.endswith('/'):

                # Continue if folder already created
                if my_bucket_object.key in created_folders:
                    continue

                # Create the folder tree and add it to the created set
                os.makedirs(path)
                created_folders.add(my_bucket_object.key)

            # Download the object if it is a file
            else:
                file_dict = {
                'file_path': path
                }

                S3Object.download(bucket_name, my_bucket_object.key, file_dict)

